/*
' Copyright (c) 2004 Tim Jacobs twjacobs@gmail.com 
'
' This program is free software; you can redistribute
' it and/or modify it under the terms of the GNU
' General Public License as published by the Free
' Software Foundation; either version 2 of the
' License, or (at your option) any later version.
'
' This program is distributed in the hope that it will
' be useful, but WITHOUT ANY WARRANTY; without even
' the implied warranty of MERCHANTABILITY or FITNESS
' FOR A PARTICULAR PURPOSE. See the GNU General Public
' License for more details.
'
' You should have received a copy of the GNU General
' Public License along with this program; if not,
' write to the Free Software Foundation, Inc., 675
' Mass Ave, Cambridge, MA 02139, USA.
*/

/*
'  clens is based on PTlens, written by Thomas Niemann
'  you can find PTLens at http://www.epaperpress.com/ptlens/
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include <getopt.h>
#include "../include/jhead.h"
#include "../include/clens.h"
#include "../include/snip_str.h"

/** \file main.c
 * \brief A program to perform lens corrections
 * \author Tim Jacobs
 * \date 2004.11.13
 *
 * clens is a command line program used to correct image distortions
 * inherent in camera lenses. It is based on PTLens written by Thomas
 * Niemann. PTLens was written for Windows, clens is an ANSI 'C' port
 * which should be able to be used on any platform.
 *
 * clens requires data from several sources to make these corrections.
 * Lens correction data is provided through a collection of files
 * referenced by one main file called "profile.txt". You must tell
 * clens where this file is. In addition to this lens calibration data,
 * clens must know which camera and lens were used to take the picture.
 * The camera used to take the picture is obtained from the EXIF data
 * found in the jpeg file itself. The lens used to take the picture must
 * be supplied by the user. 
 */
 
 

extern CamNode *pCamHdr;
extern LnsNode *pLnsHdr;
extern CoefType coef[];
extern long coefIndex;

/** \brief Data structure to hold program preferences
 */
prefType pref;
/** \brief Pointer to lens for current correction
 */
LnsNode *thisLens;
/** \brief Pointer to camera for current correction
 */
CamNode *thisCamera;
/** \brief Data structure to hold EXIF information
 */
ExifInfo exif;
/** \brief Flag set by --verbose
 */
static int verbose_flag=0;
/** \brief Flag set by --tiff and --jpeg
 */
static int format_flag=0;
/** \brief Flag set by --help
 */
static int help_flag=0;
/** \brief Flag set by --version
 */
static int version_flag=0;
/** \brief Flag set by --Lenses
 */
static int Lenses_flag=0;





/** \brief Display usage information for clens.
 * \return void.
 */
void display_help()
{
	puts("Usage: clens [OPTION]... [FILE}...");
	puts("Apply lens distortion corrections to jpeg images.");
	puts("Both barrel distortion and pincusion distortions are corrected.\n");
	puts(" -u, --unscaled\t\tdo not scale the image to remove black borders");
	puts("\t\t\tresulting from lens correction -- default is to scale the image.");
	puts(" -l, --lens\t\tidentify the lens used to take the picture.");
	puts("\t\t\tThis sets the correction coefficients -- defaults to \"Standard\"");
	puts("     --Lenses\t\tDisplay a list of supported lenses for the camera");
	puts("\t\t\tused to take each picture and leave.");
	puts(" -s, --stitcher\t\tpath to the stitcher.");
	puts("\t\t\tIf not set clens assumes it's nona.");
	puts(" -p, --profile\t\tspecify where the database is.");
	puts("\t\t\tThis should be the path to \"profile.txt\" -- no default.");
	puts("\t\t\tThis may also be set using the CLENS_PROFILE environment variable.");
	puts(" -a, --append\t\tthe string to append to each corrected file name.");
	puts("\t\t\tDefaults to \"_pt\"");
	puts(" -q, --quality\t\tsets the quality level for jpeg compression (0-100)");
	puts("\t\t\tDefaults to 90.");
	puts("     --help\t\tdisplay his help and exit.");
	puts("     --version\t\tdisplay version information and exit.");
	puts("     --tiff\t\tSet the output image format to tiff.");
	puts("     --jpeg\t\tSet the output image format to jpeg. This is the default");
}





/** \brief Display all information read from profile files.
 * \return void.
 */
void print_database()
{
	CamNode *pCurCam;
	LnsNode *pCurLns;
	GrpNode *pCurCamGrp;
	long i;
	
	
//Traverse Camera linked list
	pCurCam = pCamHdr;
	printf("\nCamera Information\n");
	while (pCurCam != NULL)
	{
		printf("%-35s %-25s", pCurCam->exifMake, pCurCam->exifModel);
		pCurCamGrp = pCurCam->group;
		while (pCurCamGrp != NULL)
		{
			printf("%-10s", pCurCamGrp->name);
			pCurCamGrp = pCurCamGrp->nextGrp;
		}
		printf("\n");
		pCurCam = pCurCam->nextCam;
	}

//Traverse coefficients array
	printf("\nCoefficients\n");
	printf("%-6s%-8s%-8s%-8s\n", "f", "a", "b", "c");
	for (i=0; i<coefIndex; i++)
	{
		printf("%3.2f %1.6f %1.6f %1.6f\n", coef[i].f, coef[i].a, coef[i].b, coef[i].c);
	}
	
//Traverse Lens linked list
	printf("\nLens Information\n");
	printf("%-50s%-10s\n", "Lens", "Group");
	pCurLns = pLnsHdr;
	while (pCurLns != NULL)
	{
		printf("%-50s%-10s\n", pCurLns->menuLens, pCurLns->group);
		pCurLns = pCurLns->nextLns;
	}
}







/** \brief main function for clens. Parses command line and controls
 *         flow of clens.
 * \param argc The argument count
 * \param argv The argument string vector
 */
int main(int argc, char **argv)
{
	char *loc;
	char inFile[256];
	char *tempBuf=NULL;
	char *outFile=NULL;
	FILE *nxtFile;
	char inLens[256];
	int rc;
	int c;
	int length;
	int option_index = 0;	//getopt_long stores the option index here.

//Set default preferences and values
	strcpy(inLens, "Standard");						//The standard camera lens
	pref.resize = 1;								//Scale the image to remove black border
	strcpy(pref.suffix, "_pt");						//The default suffix for the resulting filename
	pref.jpegQuality = 90;							//JPEG quality (0-100)
	strcpy(pref.stitcher, "nona");						//Assumes it's on the path
	pref.inputDir[0] = 0;							//Path will be specified as part of the input filenames
	pref.viewerPath[0] = 0;							//No viewer implimentation in this code yet
	pref.menuLens[0] = 0;							//No menus in this command line version of ptlens
	pref.menuMake[0] = 0;							//No menus in this command line version of ptlens
	pref.menuModel[0] = 0;							//No menus in this command line version of ptlens
	pref.format = 0;								//Default output format is jpeg
	pref.interpolation = 0;							//twj - Need to look into this value
	pref.profilePath = NULL;						//No default for profile.txt location
	pref.converterDetected = 0;						//twj - Need to look into this value

//parse command line arguments
     while (1)
	 {
		 static struct option long_options[] = {
			{"verbose", no_argument,       &verbose_flag, 1},	//Set the verbose flag
			{"help",    no_argument,       &help_flag,    1},	//Display a help message and exit
			{"version", no_argument,       &version_flag, 1},	//Display version and exit
			{"tiff",    no_argument,       &format_flag, 1},	//Set the format flag to tiff
			{"jpeg",    no_argument,       &format_flag, 0},	//Set the format flag to jpeg
			{"lens",    required_argument, 0, 'l'},				//identifies lens used to take picture
			{"Lenses",	no_argument,       &Lenses_flag, 1},	//Show all lenses for camera
			{"stitcher",required_argument, 0, 's'},				//where is stitcher
			{"profile", required_argument, 0, 'p'},				//specify database
			{"unscaled",no_argument,       0, 'u'},				//Don't resize picture to remove black edges
			{"append",  required_argument, 0, 'a'},				//file suffix for resulting image
			{"quality", required_argument, 0, 'q'},				//jpeg Q factor
			{0, 0, 0, 0}
         };
           
		c = getopt_long (argc, argv, "rl:s:p:q:a:", long_options, &option_index);
		if (c == -1)		//Detect the end of the options.
			break;
		switch (c)
		{
		case 0:				//If this option set a flag, do nothing else now.
			if (long_options[option_index].flag != 0)
				break;
			printf ("option %s", long_options[option_index].name);
			if (optarg)
				printf (" with arg %s", optarg);
			printf ("\n");
			break;
		case 'u':
			pref.resize = 0;
			break;
		case 'l':
			strcpy(inLens, optarg);
			break;
		case 's':
			strcpy(pref.stitcher, optarg);
			break;
		case 'p':
			pref.profilePath = optarg;
			break;
		case 'a':
			strcpy(pref.suffix, optarg);
			break;
		case 'q':
			pref.jpegQuality = atoi(optarg);
			break;
		case '?':			//getopt_long already printed an error message.
			break;
		default:
			abort();
		}
	 }

	 if (help_flag)
	 {
		 display_help();
		 exit(0);
	 }

	 if (version_flag)
	 {
		 puts(VERSION);
		 exit(0);
	 }
     
	 if (format_flag)
		 pref.format = 1;

//If profile is not set by command line argument,
//	then check the environment.
	 if (pref.profilePath == NULL)
	 	pref.profilePath = getenv("CLENS_PROFILE");
	 if (pref.profilePath == NULL)
	 {
		 // Specific Mandrakelinux patch, profiles are put in /usr/share/clens/profile.txt
		 // wich is set when the rpm is built
		 pref.profilePath = "/usr/share/clens/profile.txt";
	 }

//verify the existance of stitcher in current path
	 //oops no code here


//read the profile database into linked lists
	if (readDataBase() != 0)
	{
		puts ("Error reading database");
		puts ("Must specify where \"profile.txt\" is! Use the \"-p\" option");
		puts("or specify it using the CLENS_PROFILE environment variable");
		exit(1);
	}

	while (optind < argc)						//For each picture
	{
		strcpy (inFile, argv[optind++]);
		printf("%s:\n", inFile);

//verify input file exists
		nxtFile = fopen(inFile, "r");
		if (nxtFile != NULL) {					// exists, so close it
			fclose(nxtFile);
		} else {
			puts("\tFile not found");
			goto lbl_Next_Image;
		}


//simple check if the inFile is a jpeg file.
		loc = strIstr(inFile, ".jpg");			//Case insensitive compare
		length = (int)(loc - inFile);			//length used to construct new filename below
		if ( loc == NULL )
		{
			printf("\t%s does not appear to be a jpg file. It must have an extension \".jpg\"\n", inFile);
			goto lbl_Next_Image;
		}


//construct the outfile string
		outFile = (char *)malloc((strlen(inFile) +
								strlen(pref.suffix) + 2) * sizeof(char));
		strncpy(outFile, inFile, length);
		outFile[length] = '\0';
		strcat(outFile, pref.suffix);
		if (pref.format == JPEG)
			strcat(outFile, ".jpg");
		else
			strcat(outFile, ".tiff");


//read the exif data
		if (getExif(inFile,&exif) != 0)
		{
			puts("\tExif read failed");
			goto lbl_Next_Image;
		}
		if (verbose_flag)
			puts("\tExif data read from file");


//find the camera in the database
		thisCamera = findCamera(exif.make, exif.model);
		if (thisCamera == NULL)
		{
			printf("\tCamera %s not found in database.\n", exif.model);
			goto lbl_Next_Image;
		}
		if (verbose_flag)
			printf("\t%s found in database\n", exif.model);

//If user wants to print list of compatable lenses, do it and exit
		LnsNode *lenses;
		if(Lenses_flag)
		{
			printf("\tCamera: %s\n", thisCamera->menuModel);
			lenses = findLenses(thisCamera);
			while (lenses != NULL)
			{
				printf("\t%s\n", lenses->menuLens);
				lenses = lenses->nextLns;
			}
			puts("\n");
			goto lbl_Next_Image;
		}


//find a camera-compatable lens that matches the lens given on the command line
		thisLens = findLens(inLens, thisCamera);
		if (thisLens == NULL)
		{
			printf("\t%s lens not found in database.\n", inLens);
			goto lbl_Next_Image;
		}
		if (verbose_flag)
			printf("\t%s lens found in database.\n", inLens);



//prepare the stitcher script for nona or PTStitcher
		rc = prepareScript(inFile);
		if (rc != 0)
		{
			puts("\tError creating Stitcher script");
			goto lbl_Next_Image;
		}

		tempBuf = (char *)malloc((strlen(pref.stitcher) +
											strlen(outFile) + 21) * sizeof(char));
		strcpy(tempBuf, "\"");		//quote stitcher path incase it
		strcat(tempBuf, pref.stitcher); // has spaces
		strcat(tempBuf, "\"");
		strcat(tempBuf, " -o ");
		strcat(tempBuf, "\"");		//quote the output in case the
		strcat(tempBuf, outFile);	//filename has spaces
		strcat(tempBuf, "\" _script.txt");

//execute stitcher program
		if (system (tempBuf) == -1)
		{
			puts("\tError calling stitcher program - is it on your path?");
			goto lbl_Next_Image;
		}

//Copy Exif data to new file if it is jpeg
		/* there is a problem here with PTStitcher. It makes the file
		 * with the extension all caps even though this code calls for
		 * lower case. So the copyExif function doesn't find the outFile
		 * name and the exif data doesn't get copied. The nona
		 * doesn't have this problem.
		*/
		if (pref.format == JPEG)
		{
			if ((nxtFile = fopen(outFile, "r")))
			{
				fclose(nxtFile);
				rc = copyExif(inFile, outFile);
			}
			else {					//do the PTStitcher cludge
				length = strlen(outFile);
				outFile[length-3] = 'J';
				outFile[length-2] = 'P';
				outFile[length-1] = 'G';
				if ((nxtFile = fopen(outFile, "r")))
				{
					fclose(nxtFile);
					rc = copyExif(inFile, outFile);
				}
			}
		}

lbl_Next_Image:;
	}
	//All done with the images
	//Clean up temporary files and free buffers
	if (tempBuf != NULL)
		free(tempBuf);
	if (outFile != NULL)
		free(outFile);
	if((nxtFile = fopen("_thumb.jpg", "r")))
	{
		fclose(nxtFile);
		remove("_thumb.jpg");
	}
	if((nxtFile = fopen("_script.txt", "r")))
	{
		fclose(nxtFile);
		remove("_script.txt");
	}
	exit(1);
}
