/*
' Copyright (C) 2004, Tim Jacobs twjacobs@gmail.com
'
' This program is free software; you can redistribute
' it and/or modify it under the terms of the GNU
' General Public License as published by the Free
' Software Foundation; either version 2 of the
' License, or (at your option) any later version.
'
' This program is distributed in the hope that it will
' be useful, but WITHOUT ANY WARRANTY; without even
' the implied warranty of MERCHANTABILITY or FITNESS
' FOR A PARTICULAR PURPOSE. See the GNU General Public
' License for more details.
'
' You should have received a copy of the GNU General
' Public License along with this program; if not,
' write to the Free Software Foundation, Inc., 675
' Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include "../include/jhead.h"
#include "../include/clens.h"



/** \file modPT.c
 * \brief functions to manipulate lens correction coefficients
 * and create stitcher scripts
 *
 * \author Tim Jacobs (ported from Thomas Niemann's PTLens code)
 * \date 2004.11.13
 *
 * Full Description here.
 */


/** \brief Measurement of 35mm film format diagonal
 *
 * if assuming 23.3 x 35 (phil's web site) then
 *    diag35 = 42.046284
 * if assuming 36 x 24
 *    diag35 = 43.266615
 */
const double diag35 = 43.266615;
/** \brief Measurement of 35mm film format diagonal 
 */
typedef struct
{
	double hfov;
	double a;
	double b;
	double c;
	double d;
}Data1;
/** \brief Holds hfov and correction coefficients for current image
 */
Data1 coef1;

extern ExifInfo exif;
extern CamNode pCurCam;
extern prefType pref;
extern CoefType coef[];
extern LnsNode *thisLens;
extern CamNode *thisCamera;


/**
 * \brief Calculate the Horizontal Field Of View (HFOV) for
 *        a given focal length, width, height, and multiplier.
 * \return The HFOV.
 * \param foc The focal length of the lens when the picture was
 *        taken. Read from EXIF data.
 *
 *****************************************************/
double getHfov(double foc)
{
    double rad;
    double widthmm;
    double diagpx;
    double diagmm;
    double ret_val;

    // compute sensor width
    diagpx = sqrt(exif.width * exif.width + exif.height * exif.height);
    diagmm = diag35 / thisCamera->multiplier;
    widthmm = exif.width * diagmm / diagpx;
    
    rad = 2 * atan(widthmm / 2 / foc);
    ret_val = 180 * rad / (4 * atan(1));
    return(ret_val);
}



/**
 * \brief Calculate coefficient "d". Result put into the global
 *        variable coef1.d.
 *
 *****************************************************/
void coefD()
{
    double r_test[2];
    double p, r, r_fixed, fixx, test;
    int test_points, i;
    double a, b, c;
    
    if (pref.resize == 0)
    {
        coef1.d = 1 - (coef1.a + coef1.b + coef1.c);
        return;
    }
    
    // truncate black edges
    // algorithm courtesy of Paul Wilkinson, paul.wilkinson@ntlworld.com
    a = coef1.a;
    b = coef1.b;
    c = coef1.c;
    if (exif.width > exif.height)
        p = (double)(exif.width) / (double)(exif.height);
    else
        p = (double)(exif.height) / (double)(exif.width);
    
    //***************************************************
    //* Set the test point for the far corner.          *
    //***************************************************
    r_test[0] = sqrt(1 + p * p);
    test_points = 1;

    //***************************************************
    //* For non-zero values of a, there are two other   *
    //* possible test points.                           *
    //***************************************************
    if (a != 0.0)
    {
        r = (-b + sqrt(b * b - 3 * a * c)) / (3 * a);
        if (r >= 1 && r <= r_test[0])
        {
            r_test[test_points] = r;
            test_points = test_points + 1;
        }
        r = (-b - sqrt(b * b - 3 * a * c)) / (3 * a);
        if (r >= 1 && r <= r_test[0])
        {
            r_test[test_points] = r;
            test_points = test_points + 1;
        }
    }
    
    //***************************************************
    //* For zero a and non-zero b, there is one other   *
    //* possible test point.                            *
    //***************************************************
    if (a == 0 && b != 0)
    {
        r = -c / (2 * b);
        if (r >= 1 && r <= r_test[0])
        {
            r_test[test_points] = r;
            test_points = test_points + 1;
        }
    }
    
    //***************************************************
    //* Set the fixed point to Helmut's orginal point.  *
    //***************************************************
    r_fixed = 1.0;
    fixx = a + b + c;
    
    //***************************************************
    //* Test other possible fixed points.               *
    //***************************************************
    for (i = 0; i <= test_points - 1; i++)
    {
        r = r_test[i];
        test = r * (c + r * (b + r * a));
        if (test > fixx)
        {
            fixx = test;
            r_fixed = r;
        }
    }
    coef1.d = 1 - fixx;
}



/**
 * \brief Calculate the lens correction coefficients (a,b,c) for
 *        a particular focal length of a particular lens by linear
 *        interpolation of known coefficients for the lens. The 
 *        global variable "thisLens" must be pointing to the lens
 *        of interest.
 * \return 1 if unsuccessful, 0 otherwise. If successful, coef1.a,
 *         coef1.b, coef1.c, and coef.d are populated.
 *
 *****************************************************/
int getCoefs()
{
    long lb;
    long ub;
    long i;
    double ratio;
    double f;
    
    f = exif.focalLength;
    if (pref.converterDetected == 0)
        f = f * thisLens->converterFactor;
    
    lb = thisLens->coefLB;
    ub = thisLens->coefUB;

    // force focal length to be legal
    if (f < coef[lb].f)
    {
        if (f < 0.9 * coef[lb].f)
            return (1);
        f = coef[lb].f;
    }
    if (f > coef[ub].f)
    {
        if (f > 1.1 * coef[ub].f)
            return (1);
        f = coef[ub].f;
    }

    // calculate hfov
    coef1.hfov = getHfov(f);

    // calculate a,b,c
    if (lb == ub)
    {
        coef1.a = coef[lb].a;
        coef1.b = coef[lb].b;
        coef1.c = coef[lb].c;
    } else {
        i = lb;
        while (coef[i].f < f)
        {
            i = i + 1;
        }
        if (f == coef[i].f)
        {
            coef1.a = coef[i].a;
            coef1.b = coef[i].b;
            coef1.c = coef[i].c;
        } else {
            ratio = (f - coef[i - 1].f) /
            		(coef[i].f - coef[i - 1].f);
            coef1.a = coef[i - 1].a + ratio *
            		(coef[i].a - coef[i - 1].a);
            coef1.b = coef[i - 1].b + ratio *
            		(coef[i].b - coef[i - 1].b);
            coef1.c = coef[i - 1].c + ratio *
            		(coef[i].c - coef[i - 1].c);
        }
    }
    
    // multiplier correction
    if (thisLens->multiplier != thisCamera->multiplier)
    {
        double factor;
        factor = thisLens->multiplier / thisCamera->multiplier;
        coef1.a = coef1.a * factor * factor * factor;
        coef1.b = coef1.b * factor * factor;
        coef1.c = coef1.c * factor;
    }
    
    coefD();
	return(0);
}




/* twj only used by getInfo which in turn is only used to
 * display information to the user. Command line version won't
 * need this.
char *parms()
{
    char *hfov;
    char *ret_buf;
    
    parms = vbCrLf & vbCrLf & "Parameters" & vbCrLf;
    if (getCoefs() != 0)
    {
    // twj need to do some (malloc?) string copy here
        ret_buf = ret_buf & "Focal length not calibrated.";
        return(ret_buf);
    }

    if (coef1.hfov < 10.0)
    {
    // twj need to do some string copy here
        hfov = format(coef1.hfov, "0.00");
    } else {
    // need to do some string copy here
        hfov = format(coef1.hfov, "0.0");
    }

	// twj need to do some string copy here 
    parms = parms & _
        "    hfov: " & hfov & vbCrLf & _
        "    a:    " & format(coef1.a, " 0.000000;-0.000000") & vbCrLf & _
        "    b:    " & format(coef1.b, " 0.000000;-0.000000") & vbCrLf & _
        "    c:    " & format(coef1.c, " 0.000000;-0.000000") & vbCrLf & _
        "    d:    " & format(coef1.d, " 0.000000;-0.000000")
}
*/




/* twj only used by getInfo which in turn is only used to
 * display information to the user. Command line version won't
 * need this.
Private Function settings(valid As Boolean) As String
    Dim speed As String
    Dim mm As Single
    Dim t As Long
    Dim diag As Single
    Dim diag35 As Single
    Dim foc As String

    If modExif.exif.speed <= 0.5 Then
        t = 1 / modExif.exif.speed
        speed = "1/" & Trim(Str(t)) & " sec"
    Else
        speed = format(modExif.exif.speed, "0.000") & " sec"
    End If
    
    settings = settings & vbCrLf & vbCrLf & "Settings" & vbCrLf
    settings = settings & "    "
    If modExif.exif.iso <> 0 Then
        settings = settings & "ISO " & Trim(Str(modExif.exif.iso)) & ", "
    End If
    settings = settings & speed & ", f/" & format(modExif.exif.fstop, "0.0") & vbCrLf

    foc = format(modExif.exif.focalLength, "0.000")
                 
    If valid Then
        mm = modExif.exif.focalLength * pCurCam.multiplier
        If modIni.pref.converterDetected = 0 Then
            mm = mm * pCurLns.converterFactor
        End If
        If mm = modExif.exif.focalLength Then
            settings = settings & _
                "    Lens: " & Trim(Str(CLng(mm))) & "mm" & vbCrLf
        Else
   
            settings = settings & _
                "    Lens: " & Trim(Str(CLng(mm))) & "mm (" & _
                foc & " EXIF)" & vbCrLf
        End If
    Else
        settings = settings & _
                "    Lens: " & foc & " EXIF" & vbCrLf
    End If
    settings = settings & "    " & modExif.exif.dateTime
End Function
*/




/**
 * \brief Convert degrees to radians
 * \return The value in radians
 * \param x The input in degrees
 *
 *****************************************************/
double rad(double x)
{
    double pi;
    double ret_val;
    pi = 4 * atan(1);
    ret_val = x * pi / 180;
    return(ret_val);
}




/**
 * \brief Convert radians to degrees
 * \return The value in degrees
 * \param x The input in radians
 *
 *****************************************************/
double deg(double x)
{
    double pi;
    double ret_val;
    pi = 4 * atan(1);
    ret_val = 180 * x / pi;
    return(ret_val);
}




/**
 * \brief Write the script the stitching program will use to
 *        warp the image (ie remove the lens distortion).
 * \return 0 if successful, 3 if there is a problem getting the
 *         coefficients, 5 if all coefficients are 0, 4 if unable
 *         to open the script file.
 * \param inFile The file name for the input image
 *
 *****************************************************/
int prepareScript(char *inFile)
{
    char *scriptFile = "_script.txt";
    char buf[2048];
    char tmp_buf[512];
    double h;
    FILE *file;

/*twj Moved to main
    if (libGetExif(inFile, &exif) != 0)
        return(1);
    
    if (validCamera() == 0)
        return(2);

    // if stitching, and newFile exists, ignore

	outPath = (char *)malloc((strlen(pref.inputDir) +
						strlen(outFile) +
						2) * sizeof(char));
	strcpy(outPath, pref.inputDir);
	strcat(outPath, "/");
	strcat(outPath, outFile);
    If FileExists(outPath) Then Exit Function

*/
   
    // determine script
    if (getCoefs() != 0)
    {
        return(3);
    }
    
    if (coef1.a == 0 && coef1.b == 0 && coef1.c == 0)
    {
        return(5);
    }

	/*twj moved to main()
	cmd = (char *)malloc((strlen(pref.stitcher) +
						strlen(scriptFile) +
						strlen(outPath) +
						6) * sizeof(char));	//6 for spaces, "-o" and null
	strcpy(cmd, pref.stitcher);
	strcat(cmd, " ");
	strcat(cmd, scriptFile);
	strcat(cmd, " -o ");
	strcat(cmd, outPath);
	

	//Comment at the top of script file
    strcpy(buf, "# ");
    strcat(buf, cmd);
    strcat(buf, "\n");
	*/

    // p-line
    // twj don't know why u20 (ie feather 20 pixels)... so I took it out
    sprintf(tmp_buf, "p w%d h%d f0 ", exif.width, exif.height);
    strcpy(buf, tmp_buf);

	//Output resized or not depending of preferences
    if( pref.resize == 1)
    {
        h = deg(2 * atan((coef1.a + coef1.b + coef1.c + coef1.d) *
        		tan(rad(coef1.hfov) / 2)));
        sprintf(tmp_buf, "%.5f",h);
        strcat(buf, "v");
        strcat(buf, tmp_buf);
    } else {
    	sprintf(tmp_buf, "%.5f", coef1.hfov);
    	strcat(buf, "v");
    	strcat(buf, tmp_buf);
    }

	//Output tiff or jpeg depending of preferences
	//If jpeg, quality level is again via preferences
    if (pref.format == TIFF)
    {
    	strcat(buf, " n\"TIFF\"\n");
    } else {
    	sprintf(tmp_buf, " n\"JPEG g0 q%d \"\n", pref.jpegQuality);
    	strcat(buf, tmp_buf);
    }
    
    sprintf(tmp_buf, "m g1 i%d\n", pref.interpolation);
    strcat(buf, tmp_buf);
    sprintf(tmp_buf, "o w%d h%d f0 y0 r0 p0 v%.5f a%f b%f c%f d0 e0 n\"%s\"",
    			exif.width, exif.height,
				coef1.hfov, coef1.a, coef1.b, coef1.c,
				inFile);
    strcat(buf, tmp_buf);

//Write out the script file
	file = fopen(scriptFile, "w");
	if (file == NULL)
	{
		printf("Unable to open script file.\n");
		return(4);
	}
    fprintf(file,"%s",buf);
    fclose(file);
    return(0);
}

/*twj Moved to main
int validCamera()
{
	int ret_value=0;
	if ((pCurCam.exifMake == exif.make) &&
			(pCurCam.exifModel == exif.model))
	{
		ret_value=1;
	}
	return(ret_value);
}
*/


/* twj only used to display information to the user
 * Command line version won't need this.
Function getInfo(fileName As String, info As String) As Long
    Dim vc As Boolean
    
    getInfo = getExif(fileName, 1)
    If getInfo <> 0 Then Exit Function
    
    info = modExif.exif.make & vbCrLf & modExif.exif.model & vbCrLf
    info = info & Trim(Str(modExif.exif.width)) & " x " & Trim(Str(modExif.exif.height))
    
    vc = validCamera
    
    If vc Then
        info = info & settings(1)
        info = info & parms()
    Else
        info = info & settings(0)
        info = info & vbCrLf & vbCrLf & "Parameters" & vbCrLf
        info = info & "    EXIF info doesn't match setup." & vbCrLf
        If pCurCam.exifMake <> modExif.exif.make Then
            info = info & "    Wrong camera make." & vbCrLf
        End If
        If pCurCam.exifModel <> modExif.exif.model Then
            info = info & "    Wrong camera model." & vbCrLf
        End If
    End If
End Function
*/
